;*=====================================================================*/
;*    .../prgm/project/bigloo/api/multimedia/src/Llib/mpg123.scm       */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Tue Jul 10 10:45:58 2007                          */
;*    Last change :  Sat Nov 24 20:57:59 2007 (serrano)                */
;*    Copyright   :  2007 Manuel Serrano                               */
;*    -------------------------------------------------------------    */
;*    The MPG123 Bigloo binding                                        */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module __multimedia-mpg123
   
   (import __multimedia-music)
   
   (export (class mpg123::music
	      (make-thread::procedure read-only)
	      (path::bstring read-only (default "mpg123"))
	      (args::pair-nil read-only (default '("--remote")))
	      (%thread::obj (default #f))
	      (%playlist::pair-nil (default '()))
	      (%playlistid::int (default 0))
	      (%index::int (default 0))
	      (%process (default #f))
	      (%condv::condvar read-only (default (make-condition-variable)))
	      (%mutex2::mutex read-only (default (make-mutex "mpg123-mutex2")))
	      (%mutex::mutex read-only (default (make-mutex "mpg123-mutex")))
	      (%error (default #unspecified))
	      (%state (default 'stop))
	      (%position::int (default 0))
	      (%length::int (default 0))
	      (%volume::int (default 100))
	      (%bitrate::int (default 0))
	      (%khz::int (default 0)))))

;*---------------------------------------------------------------------*/
;*    mpg123-acknowledge ...                                           */
;*---------------------------------------------------------------------*/
(define (mpg123-result-acknowledge) "@R MPG123")

(define (mpg123-command-seek) "JUMP")
(define (mpg123-command-load) "LOAD")
(define (mpg123-command-loadpaused) "LOADPAUSED")
(define (mpg123-command-pause) "PAUSE")
(define (mpg123-command-stop) "STOP")
(define (mpg123-command-volume) "VOLUME")

;*---------------------------------------------------------------------*/
;*    mpg123-exec ...                                                  */
;*---------------------------------------------------------------------*/
(define (mpg123-exec proc command #!optional arg)
   (let ((p (process-input-port proc)))
      (display command p)
      (when arg
	 (display " " p) 
	 (display arg p))
      (newline p)
      (flush-output-port p)))

;*---------------------------------------------------------------------*/
;*    number-grammar ...                                               */
;*---------------------------------------------------------------------*/
(define number-grammar
   (regular-grammar ()
      (#\space (ignore))
      ((+ digit) (the-fixnum))
      ((: (+ digit) #\. (* digit)) (the-fixnum))
      (else (the-failure))))
	 
;*---------------------------------------------------------------------*/
;*    mpg123-grammar ...                                               */
;*---------------------------------------------------------------------*/
(define mpg123-grammar
   (regular-grammar ((int (+ digit))
		     (float (: (+ digit) #\. (* digit)))
		     mpg123 armed)
      ("@E"
       (let ((reason (read-line (the-port))))
	  (with-access::mpg123 mpg123 (%mutex %condv %state %error)
	     (mutex-lock! %mutex)
	     (set! %state 'error)
	     (set! %error reason)
	     (condition-variable-signal! %condv)
	     (mutex-unlock! %mutex))
	  (ignore)))
      ((: "@F")
       ;; Status message during playing
       (let* ((a (read/rp number-grammar (the-port)))
	      (b (read/rp number-grammar (the-port)))
	      (c (read/rp number-grammar (the-port)))
	      (d (read/rp number-grammar (the-port))))
	  (with-access::mpg123 mpg123 (%mutex %state %position %length)
	     (mutex-lock! %mutex)
	     (set! %state 'play)
	     (set! %position c)
	     (set! %length (+fx c d))
	     (cond
		((<=fx d 0)
		 (when armed
		    (set! armed #f)
		    ;; start playing next song
		    (with-access::mpg123 mpg123 (%process %index %playlist)
		       (when (<fx %index (length %playlist))
			  (set-song! mpg123 %index)
			  (mpg123-exec %process (mpg123-command-pause))))))
		((not armed)
		 ;; we have frames to play
		 (set! armed #t)))
	     (mutex-unlock! %mutex))
	  (ignore)))
      ((: "@S ")
       ;; Stream info at beginning of playback
       (let* ((mpeg-type (read (the-port)))
	      (layer (read (the-port)))
	      (sampling (read (the-port)))
	      (mode (read (the-port)))
	      (mode-ext (read (the-port)))
	      (framesize (read (the-port)))
	      (stereo (read (the-port)))
	      (copyright (read (the-port)))
	      (err (read (the-port)))
	      (emphasis (read (the-port)))
	      (bitrate (read (the-port))))
	  (read-line (the-port))
	  (with-access::mpg123 mpg123 (%mutex %state %khz %bitrate)
	     (mutex-lock! %mutex)
	     (set! %khz sampling)
	     (set! %bitrate bitrate)
	     (mutex-unlock! %mutex)
	     (ignore))))
      ((: "@P " digit)
       ;; Playing status
       (with-access::mpg123 mpg123 (%mutex %condv %state)
	  (mutex-lock! %mutex)
	  (case (integer->char (the-byte-ref 3))
	     ((#\0)
	      (set! %state 'stop))
	     ((#\1)
	      (set! %state 'pause))
	     ((#\2)
	      (set! %state 'play)))
	  (mutex-unlock! %mutex))
       (ignore))
      ((: "@I" (+ all))
       (ignore))
      ((: "@" (or (in "JV") "RVA" "bass") (+ all))
       (ignore))
      ((: #\@ (in alpha))
       (tprint (the-string) (read-line (the-port)))
       (ignore))
      ((+ #\Newline)
       (ignore))
      (else
       (let ((c (the-failure)))
	  (unless (eof-object? c)
	     (raise
	      (instantiate::&io-parse-error
		 (proc 'mpg123)
		 (msg "Illegal response")
		 (obj (format "{~a}~a" c (read-line (the-port)))))))))))
	 
;*---------------------------------------------------------------------*/
;*    mpg123-thread ...                                                */
;*---------------------------------------------------------------------*/
(define (mpg123-thread mpg123)
   (with-access::mpg123 mpg123 (%condv %mutex %process)
      (let ((p (process-output-port %process))
	    (buf (make-string 256)))
	 (read/rp mpg123-grammar p mpg123 #t))))
   
;*---------------------------------------------------------------------*/
;*    mpg123-start ...                                                 */
;*---------------------------------------------------------------------*/
(define (mpg123-start path args)
   (let ((proc (apply run-process
		      path
		      input: pipe: output: pipe: error: "/dev/null"
		      wait: #f fork: #t
		      args)))
      (if (not (process-alive? proc))
	  (raise
	   (instantiate::&io-error
	      (proc 'mpg123-start)
	      (msg "Can't start process")
	      (obj (format "~a ~a" path args))))
	  (let ((l (read-line (process-output-port proc))))
	     (if (not (substring-at? l (mpg123-result-acknowledge) 0))
		 (raise
		  (instantiate::&io-parse-error
		     (proc 'mpg123-start)
		     (msg "Illegal MPG123 acknowledge")
		     (obj l)))
		 proc)))))

;*---------------------------------------------------------------------*/
;*    mpg123-connect! ...                                              */
;*---------------------------------------------------------------------*/
(define (mpg123-connect! mpg123::mpg123)
   (with-access::mpg123 mpg123 (make-thread %thread %process path args)
      ;; start a new mpg213 process
      (unless (and (process? %process) (process-alive? %process))
	 (set! %process (mpg123-start path args)))
      ;; start a new thread for driving this mpg123 instance
      (unless %thread
	 (set! %thread (make-thread
			(lambda ()
			   (mpg123-thread mpg123)))))))

;*---------------------------------------------------------------------*/
;*    music-close ::mpg123 ...                                         */
;*---------------------------------------------------------------------*/
(define-method (music-close mpg123::mpg123)
   (with-access::mpg123 mpg123 (%mutex %process)
      (with-lock %mutex
	 (lambda ()
	    (when (process? %process)
	       (process-kill %process)
	       (set! %process #f))))))

;*---------------------------------------------------------------------*/
;*    music-close ::mpg123 ...                                         */
;*---------------------------------------------------------------------*/
(define-method (music-close mpg123::mpg123)
   (with-access::mpg123 mpg123 (%mutex %process)
      (eq? %process #f)))

;*---------------------------------------------------------------------*/
;*    music-reset! ::mpg123 ...                                        */
;*---------------------------------------------------------------------*/
(define-method (music-reset! mpg123::mpg123)
   (music-close mpg123))

;*---------------------------------------------------------------------*/
;*    music-abort ::mpg123 ...                                         */
;*---------------------------------------------------------------------*/
(define-method (music-abort mpg123::mpg123)
   (with-access::mpg123 mpg123 (%mutex %process %state)
      (with-lock %mutex
	 (lambda ()
	    (set! %state 'stop)
	    (mpg123-connect! mpg123)
	    (mpg123-exec %process (mpg123-command-stop))))))

;*---------------------------------------------------------------------*/
;*    music-playlist-get ::mpg123 ...                                  */
;*---------------------------------------------------------------------*/
(define-method (music-playlist-get mpg123::mpg123)
   (with-access::mpg123 mpg123 (%playlist)
      %playlist))

;*---------------------------------------------------------------------*/
;*    music-playlist-add! ::mpg123 ...                                 */
;*---------------------------------------------------------------------*/
(define-method (music-playlist-add! mpg123::mpg123 s)
   (with-access::mpg123 mpg123 (%mutex %playlist %playlistid)
      (with-lock %mutex
	 (lambda ()
	    (set! %playlistid (+fx 1 %playlistid))
	    (set! %playlist (append %playlist (list s)))))))

;*---------------------------------------------------------------------*/
;*    music-playlist-delete! ::mpg123 ...                              */
;*---------------------------------------------------------------------*/
(define-method (music-playlist-delete! mpg123::mpg123 s)
   (with-access::mpg123 mpg123 (%mutex %playlist)
      (with-lock %mutex
	 (lambda ()
	    (set! %playlist (delete! s %playlist string=?))))))

;*---------------------------------------------------------------------*/
;*    music-playlist-clear! ::mpg123 ...                               */
;*---------------------------------------------------------------------*/
(define-method (music-playlist-clear! mpg123::mpg123)
   (with-access::mpg123 mpg123 (%mutex %playlist %index)
      (with-lock %mutex
	 (lambda ()
	    (set! %index 0)
	    (set! %playlist '())))))

;*---------------------------------------------------------------------*/
;*    set-song! ...                                                    */
;*---------------------------------------------------------------------*/
(define (set-song! mpg123 i)
   (with-access::mpg123 mpg123 (%process %index %playlist)
      (if (or (<fx i 0) (>=fx i (length %playlist)))
	  (raise
	   (instantiate::&io-error
	      (proc '|set-song ::mpg123|)
	      (msg (format "No such soung: ~a" i))
	      (obj %playlist)))
	  (let ((m (list-ref %playlist i)))
	     (set! %index (+fx i 1))
	     (mpg123-exec %process (mpg123-command-loadpaused) m)))))

;*---------------------------------------------------------------------*/
;*    music-play ::mpg123 ...                                          */
;*---------------------------------------------------------------------*/
(define-method (music-play mpg123::mpg123 . song)
   (with-access::mpg123 mpg123 (%mutex %process %index)
      (with-lock %mutex
	 (lambda ()
	    (mpg123-connect! mpg123)
	    (if (pair? song)
		(if (not (integer? (car song)))
		    (bigloo-type-error '|music-play ::mpg123| 'int (car song))
		    (set-song! mpg123 (car song)))
		(set-song! mpg123 %index))
	    (mpg123-exec %process (mpg123-command-pause))))))

;*---------------------------------------------------------------------*/
;*    music-seek ::mpg123 ...                                          */
;*---------------------------------------------------------------------*/
(define-method (music-seek mpg123::mpg123 pos . song)
   (with-access::mpg123 mpg123 (%mutex %process)
      (with-lock %mutex
	 (lambda ()
	    (mpg123-connect! mpg123)
	    (if (pair? song)
		(if (not (integer? (car song)))
		    (bigloo-type-error '|music-seek ::mpg123| 'int (car song))
		    (begin
		       (set-song! mpg123 (car song))
		       (mpg123-exec %process (mpg123-command-seek)
				    (format "~as" pos))
		       (mpg123-exec %process (mpg123-command-pause))))
		(mpg123-exec %process (mpg123-command-seek)
			     (format "~as" pos)))))))

;*---------------------------------------------------------------------*/
;*    music-stop ::mpg123 ...                                          */
;*---------------------------------------------------------------------*/
(define-method (music-stop mpg123::mpg123)
   (with-access::mpg123 mpg123 (%mutex %process)
      (with-lock %mutex
	 (lambda ()
	    (mpg123-connect! mpg123)
	    (mpg123-exec %process (mpg123-command-stop))))))

;*---------------------------------------------------------------------*/
;*    music-pause ::mpg123 ...                                         */
;*---------------------------------------------------------------------*/
(define-method (music-pause mpg123::mpg123)
   (with-access::mpg123 mpg123 (%mutex %process)
      (with-lock %mutex
	 (lambda ()
	    (mpg123-connect! mpg123)
	    (mpg123-exec %process (mpg123-command-pause))))))

;*---------------------------------------------------------------------*/
;*    music-next ::mpg123 ...                                          */
;*---------------------------------------------------------------------*/
(define-method (music-next mpg123::mpg123)
   (with-access::mpg123 mpg123 (%mutex %process %index %playlist)
      (with-lock %mutex
	 (lambda ()
	    (if (>=fx %index (length %playlist))
		(raise
		 (instantiate::&io-error
		    (proc '|music-next ::mpg123|)
		    (msg "No next soung")
		    (obj %index)))
		(let ((m (list-ref %playlist %index)))
		   (set! %index (+fx 1 %index))
		   (mpg123-connect! mpg123)
		   (mpg123-exec %process (mpg123-command-load) m)))))))

;*---------------------------------------------------------------------*/
;*    music-prev ::mpg123 ...                                          */
;*---------------------------------------------------------------------*/
(define-method (music-prev mpg123::mpg123)
   (with-access::mpg123 mpg123 (%mutex %process %index %playlist)
      (with-lock %mutex
	 (lambda ()
	    (if (or (=fx %index 0) (null? %playlist))
		(raise
		 (instantiate::&io-error
		    (proc '|music-prev ::mpg123|)
		    (msg "No previous soung")
		    (obj %index)))
		(let ((m (list-ref %playlist (-fx %index 1))))
		   (mpg123-connect! mpg123)
		   (mpg123-exec %process (mpg123-command-load) m)))))))

;*---------------------------------------------------------------------*/
;*    music-properties ::mpg123 ...                                    */
;*---------------------------------------------------------------------*/
(define-method (music-properties mpg123::mpg123)
   (with-access::mpg123 mpg123 (%mutex
				%index %playlist
				%state %index %position %length
				%volume %error %bitrate %khz)
      (with-lock %mutex
	 (lambda ()
	    (list (cons :state %state)
		  (cons :playlist (length %playlist))
		  (cons :song (-fx %index 1))
		  (cons :pos %position)
		  (cons :len %length)
		  (cons :vol %volume)
		  (cons :err %error)
		  (cons :bitrate %bitrate)
		  (cons :khz %khz))))))

;*---------------------------------------------------------------------*/
;*    music-info ::mpg123 ...                                          */
;*---------------------------------------------------------------------*/
(define-method (music-info mpg123::mpg123)
   (with-access::mpg123 mpg123 (%mutex
				%index %playlistid
				%state %index %position %length
				%volume %error %bitrate %khz)
      (with-lock %mutex
	 (lambda ()
	    (values %state
		    %playlistid
		    (-fx %index 1)
		    %position
		    %length
		    %volume
		    %error
		    %bitrate
		    %khz)))))

;*---------------------------------------------------------------------*/
;*    music-song ::mpg123 ...                                          */
;*---------------------------------------------------------------------*/
(define-method (music-song mpg123::mpg123)
   (with-access::mpg123 mpg123 (%mutex %process %index %playlist)
      (with-lock %mutex
	 (lambda ()
	    (when (pair? %playlist)
	       (list-ref %playlist (-fx %index 1)))))))

;*---------------------------------------------------------------------*/
;*    music-time ::mpg123 ...                                          */
;*---------------------------------------------------------------------*/
(define-method (music-time mpg123::mpg123)
   (with-access::mpg123 mpg123 (%mutex %process %position)
      (with-lock %mutex
	 (lambda ()
	    %position))))

;*---------------------------------------------------------------------*/
;*    music-volume-get ::mpg123 ...                                    */
;*---------------------------------------------------------------------*/
(define-method (music-volume-get mpg123::mpg123)
   (with-access::mpg123 mpg123 (%volume)
      %volume))

;*---------------------------------------------------------------------*/
;*    music-volume-set! ::mpg123 ...                                   */
;*---------------------------------------------------------------------*/
(define-method (music-volume-set! mpg123::mpg123 v)
   (with-access::mpg123 mpg123 (%mutex %process %volume)
      (with-lock %mutex
	 (lambda ()
	    (set! %volume v)
	    (mpg123-connect! mpg123)
	    (mpg123-exec %process (mpg123-command-volume) v)))))






   

