;*=====================================================================*/
;*    serrano/prgm/project/bigloo/runtime/Eval/library.scm             */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Thu Jun 23 15:31:39 2005                          */
;*    Last change :  Wed Nov 14 15:06:28 2007 (serrano)                */
;*    Copyright   :  2005-07 Manuel Serrano                            */
;*    -------------------------------------------------------------    */
;*    The library-load facility                                        */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module __library
   
   (import __error
	   __thread
	   __type
	   __bigloo
	   __configure
	   __param
	   __eval
	   __r5_control_features_6_4
	   __everror)
   
   (use     __tvector
	    __bexit
	    __os
	    __foreign
	    __evenv
	    __evmodule
	    
	    __r4_numbers_6_5_fixnum
	    __r4_numbers_6_5_flonum
	    __r4_booleans_6_1
	    __r4_symbols_6_4
	    __r4_vectors_6_8
	    __r4_control_features_6_9
	    __r4_pairs_and_lists_6_3
	    __r4_characters_6_6
	    __r4_equivalence_6_2
	    __r4_strings_6_7
	    __r4_ports_6_10_1
	    __r4_ports_6_10_1
	    __r4_output_6_10_3
	    __r4_input_6_10_2)

   (export  (library-translation-table::pair-nil)
	    (library-dlopen-init-table::pair-nil)
	    (library-translation-table-add! ::symbol ::bstring . ::obj)
	    (library-file-name::bstring ::symbol ::bstring ::symbol)
	    (library-load ::obj . opt)))

;*---------------------------------------------------------------------*/
;*    *transtable-mutex* ...                                           */
;*---------------------------------------------------------------------*/
(define *transtable-mutex* (make-mutex 'library))

;*---------------------------------------------------------------------*/
;*    *library-transtable* ...                                         */
;*---------------------------------------------------------------------*/
(define *library-transtable* '())

;*---------------------------------------------------------------------*/
;*    library-translation-table ...                                    */
;*---------------------------------------------------------------------*/
(define (library-translation-table)
   *library-transtable*)

;*---------------------------------------------------------------------*/
;*    *library-dlopen-init* ...                                        */
;*---------------------------------------------------------------------*/
(define *library-dlopen-init* '())

;*---------------------------------------------------------------------*/
;*    library-dlopen-init-table ...                                    */
;*---------------------------------------------------------------------*/
(define (library-dlopen-init-table)
   *library-dlopen-init*)

;*---------------------------------------------------------------------*/
;*    library-translation-table-add! ...                               */
;*---------------------------------------------------------------------*/
(define (library-translation-table-add! name translation . opt)
   (let ((version (bigloo-config 'release-number))
	 (init name))
      (let loop ((opt opt))
	 (when (pair? opt)
	    (cond
	       ((string? (car opt))
		(set! version (car opt))
		(loop (cdr opt)))
	       ((not (car opt))
		(set! version #f)
		(loop (cdr opt)))
	       ((eq? (car opt) :dlopen-init)
		(cond
		   ((null? (cdr opt))
		    (error 'library-translation-table-add!
			   "Illegal :dlopen-init argument"
			   opt))
		   ((not (string? (cadr opt)))
		    (error 'library-translation-table-add!
			   "Illegal :dlopen-init value"
			   opt))
		   (else
		    (set! init (cadr opt))
		    (loop (cddr opt)))))
	       (else
		(error 'library-translation-table-add!
		       "Illegal argument"
		       opt)))))
      (mutex-lock! *transtable-mutex*)
      (set! *library-transtable*
	    (cons (cons name (cons translation version)) *library-transtable*))
      (when init
	 (set! *library-dlopen-init*
	       (cons (cons name init) *library-dlopen-init*)))
      (mutex-unlock! *transtable-mutex*)))

;*---------------------------------------------------------------------*/
;*    library-init-file ...                                            */
;*---------------------------------------------------------------------*/
(define (library-init-file lib)
   (string-append (symbol->string lib) ".init"))

;*---------------------------------------------------------------------*/
;*    untranslate-library-name ...                                     */
;*---------------------------------------------------------------------*/
(define (untranslate-library-name library::symbol)
   (let ((trans (assq library (library-translation-table))))
      (if (pair? trans)
	  (values (cadr trans) (cddr trans))
	  (values (symbol->string library) (bigloo-config 'release-number)))))
   
;*---------------------------------------------------------------------*/
;*    library-file-name ...                                            */
;*---------------------------------------------------------------------*/
(define (library-file-name library suffix backend)
   (define (forge-name base suffix version)
      (cond
	 ((not version)
	  (string-append base suffix))
	 ((string? version)
	  (string-append base suffix "-" version))
	 (else
	  (error 'library-file-name "Illegal version" version))))
   (multiple-value-bind (base version)
      (untranslate-library-name library)
      (case backend
	 ((bigloo-c)
	  (cond
	     ((or (string=? (os-class) "unix")
		  (string=? (os-class) "mingw"))
	      (forge-name base suffix version))
	     ((string=? (os-class) "win32")   
	      (string-append base suffix))
	     (else
	      (error 'library-file-name "Unknown os" (os-class)))))
	 ((bigloo-jvm)
	  (forge-name base suffix version))
	 ((bigloo-.net bigloo.net)
	  (forge-name base suffix version))
	 (else
	  (error 'library-file-name "Illegal backend" backend)))))

;*---------------------------------------------------------------------*/
;*    mangle ...                                                       */
;*---------------------------------------------------------------------*/
(define (mangle name)
   (if (bigloo-need-mangling? name)
       (bigloo-mangle name)
       name))

;*---------------------------------------------------------------------*/
;*    library-load ...                                                 */
;*---------------------------------------------------------------------*/
(define (library-load lib . path)
   (let ((mod (eval-module)))
      ($eval-module-set! (interaction-environment))
      (unwind-protect
	 (if (string? lib)
	     (dynamic-load lib)
	     (let* ((path (if (pair? path)
			      path
			      (let ((venv (getenv "BIGLOOLIB")))
				 (if (not venv)
				     (bigloo-library-path)
				     (cons "." (unix-path->list venv))))))
		    (init (find-file/path (library-init-file lib) path))
		    (be (cond-expand
			   (bigloo-c 'bigloo-c)
			   (bigloo-jvm 'bigloo-jvm)
			   (bigloo-.net 'bigloo.net))))
		(when init (loadq init))
		(let* ((n (make-shared-lib-name
			   (library-file-name lib "" be) be))
		       (ns (make-shared-lib-name
			    (library-file-name lib "_s" be) be))
		       (ne (make-shared-lib-name
			    (library-file-name lib "_e" be) be))
		       (rsc (let ((p (string-append "/resource/bigloo/"
						    (symbol->string lib)
						    "/make-lib.class")))
			       ;; JVM supports fake file system in the JAR
			       ;; file. In Bigloo it is a sub-directory of
			       ;; /resource inside ressources, we always search
			       ;; for a class named make-lib. This means that
			       ;; all JVM libraries must be provided with
			       ;; such a class.
			       (and (file-exists? p) p)))
		       (libs (find-file/path ns path))
		       (libe (find-file/path ne path))
		       (name (symbol->string lib))
		       (dlopen-init (assoc name (library-dlopen-init-table)))
		       (init_s (mangle
				(if (pair? dlopen-init)
				    (string-append dlopen-init "_s")
				    (string-append name "_s"))))
		       (init_e (mangle
				(if (pair? dlopen-init)
				    (string-append dlopen-init "_e")
				    (string-append name "_e")))))
		   (cond
		      ((and (not (string? rsc)) (not (string? libs)))
		       (error 'library-load
			      (format "Can't find library `~a' (`~a')" lib ns)
			      path))
		      ((not (string? libe))
		       (cond-expand
			  ((not bigloo-jvm)
			   (evmeaning-warning
			    #f
			    'library-load
			    (format
			     "Can't find _e library `~a' (`~a') in path "
			     lib ne)
			    path)))
		       (if (string? libs)
			   (dynamic-load libs init_s)
			   (dynamic-load rsc init_e)))
		      (else
		       (if (string? libs)
			   (dynamic-load libs init_s)
			   (dynamic-load rsc init_e))
		       (dynamic-load libe init_e))))))
	 ($eval-module-set! mod))))
